/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 */

package usercontract

import (
	"crypto/sha256"
	"crypto/x509"
	"encoding/hex"
	contractapi "git.huawei.com/poissonsearch/wienerchain/contract/docker-container/contract-go/contractapi"

	"encoding/json"
	"encoding/pem"
	"fmt"
	"strconv"
	"strings"

	"github.com/pkg/errors"
)

const (
	Len256AddrToLen160Addr = 12
	indexName              = "color/size"
)

type example struct {
}

func NewSmartContract() contractapi.Contract {
	return &example{}
}

func (e *example) Init(_ contractapi.ContractStub) ([]byte, error) {
	fmt.Println("Init")
	return nil, nil
}

func (e *example) Invoke(stub contractapi.ContractStub) ([]byte, error) {
	funcName := stub.FuncName()
	args := stub.Parameters()

	// 针对简单场景，键值的操作
	switch funcName {
	case "insert":
		return insert(stub, args)
	case "query":
		return query(stub, args)
	case "delete":
		return del(stub, args)
	case "queryKeyHistory":
		return queryKeyHistory(stub, args)
	case "queryInRange":
		return queryInRange(stub, args)
	case "queryInRangeMatchCondition":
		return queryInRangeMatchCondition(stub, args)
	case "GetAddress":
		creatorBytes, _ := stub.GetPemCert()
		callerAddr, _ := CertToAddr(creatorBytes)
		sender, _ := ConvAddress(callerAddr)
		return []byte(hex.EncodeToString(sender[:])), nil
	case "GetCreator":
		creator := stub.GetOrg()
		return []byte(creator[0]), nil
	}

	// 针对复杂场景，弹珠游戏的操作
	switch funcName {
	case "insertMarble":
		return insertMarble(stub, args)
	case "queryMarble":
		return queryMarble(stub, args)
	case "deleteMarble":
		return deleteMarble(stub, args)
	case "queryMarbleKeyHistory":
		return queryMarbleKeyHistory(stub, args)
	case "queryMarbleInRange":
		return queryMarbleInRange(stub, args)
	// 复合索引的操作，可以实现高性能查询
	case "insertMarbleAndIndex":
		return insertMarbleAndIndex(stub, args)
	case "queryMarblesByIndex":
		return queryMarblesByIndex(stub, args)
	case "transferMarblesByIndex":
		return transferMarblesByIndex(stub, args)
	case "deleteMarbleAndIndex":
		return deleteMarbleAndIndex(stub, args)
	}

	return nil, errors.Errorf("func name is not correct, the function name is %s ", funcName)
}

// 插入一个普通key-value
func insert(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 2 {
		return nil, errors.New("the argNum for insert is not correct, expected 2")
	}
	keyStr := string(args[0])
	value := args[1]

	err := stub.PutKV(keyStr, value)
	if err != nil {
		return nil, err
	}
	return nil, nil
}

// 查询一个key-value
func query(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for query is not correct, expected 1")
	}

	key := args[0]
	value, err := stub.GetKV(string(key))
	if err != nil {
		errInfo := fmt.Sprintf("get the key: %s failed", key)
		return nil, errors.New(errInfo)
	}

	return value, nil
}

// 删除一个key-value
func del(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for del is not correct, expected 1")
	}

	key := args[0]
	value, err := stub.GetKV(string(key))
	if err != nil {
		errInfo := fmt.Sprintf("get the key: %s failed", key)
		return nil, errors.New(errInfo)
	}
	if value == nil {
		return nil, errors.New("the key to be delete is not exist")
	}
	err = stub.DelKV(string(key))
	if err != nil {
		return nil, err
	}

	return nil, nil
}

type keyHistory struct {
	Value     string
	BlockNum  uint64
	TxNum     int32
	TxHash    []byte
	IsDeleted bool
	Timestamp uint64
}

// 查询某个key的历史
func queryKeyHistory(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for queryKeyHistory is not correct, expected 1")
	}

	key := string(args[0])

	iterator, err := stub.GetKeyHistoryIterator(key)
	if err != nil {
		return nil, err
	}
	defer iterator.Close()

	var historyArray []keyHistory
	for iterator.Next() {
		var history keyHistory
		history.Value = string(iterator.Value())
		history.TxHash = iterator.TxHash()
		history.BlockNum, history.TxNum = iterator.Version()
		history.IsDeleted = iterator.IsDeleted()
		history.Timestamp = iterator.Timestamp()
		historyArray = append(historyArray, history)
	}

	historyMapBytes, err := json.Marshal(historyArray)
	if err != nil {
		return nil, err
	}

	return historyMapBytes, nil
}

// 查询指定范围的记录
func queryInRange(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 2 {
		return nil, errors.New("the argNum for queryInRange is not correct, expected 2")
	}

	begin, end := string(args[0]), string(args[1])

	iterator, err := stub.GetIterator(begin, end)
	if err != nil {
		return nil, err
	}
	defer iterator.Close()

	var rangeArray []map[string]string
	var count = 0
	for iterator.Next() {
		key := iterator.Key()
		value := string(iterator.Value())
		rangeArray = append(rangeArray, map[string]string{key: value})
		count++
	}
	rangeArrayBytes, err := json.Marshal(rangeArray)
	if err != nil {
		return nil, err
	}

	return rangeArrayBytes, nil
}

// 在指定key的范围内查询满足条件的记录
func queryInRangeMatchCondition(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 4 {
		return nil, errors.New("the argNum for queryInRangeMatchCondition is not correct, expected 4")
	}
	begin, end := string(args[0]), string(args[1])
	judge, v := string(args[2]), string(args[3])

	iterator, err := stub.GetIterator(begin, end)
	if err != nil {
		return nil, err
	}
	defer iterator.Close()
	var rangeArray []map[string]string
	for iterator.Next() {
		key := iterator.Key()
		value := string(iterator.Value())
		switch judge {
		case ">":
			if value > v {
				rangeArray = append(rangeArray, map[string]string{key: value})
			}
		case "<":
			if value < v {
				rangeArray = append(rangeArray, map[string]string{key: value})
			}
		case "=":
			if value == v {
				rangeArray = append(rangeArray, map[string]string{key: value})
			}
		case "!=":
			if value != v {
				rangeArray = append(rangeArray, map[string]string{key: value})
			}
		default:
			continue
		}
	}
	rangeArrayBytes, err := json.Marshal(rangeArray)
	if err != nil {
		return nil, err
	}

	return rangeArrayBytes, nil
}

func CertToAddr(creator []byte) ([]byte, error) {
	bl, _ := pem.Decode(creator)
	if bl == nil {
		return nil, errors.New("no pem data found")
	}

	cert, err := x509.ParseCertificate(bl.Bytes)
	if err != nil {
		return nil, errors.New(err.Error() + " failed to parse certificate")
	}

	pubkeyBytes, err := x509.MarshalPKIXPublicKey(cert.PublicKey)
	if err != nil {
		return nil, errors.New(err.Error() + " unable to marshal public key")
	}

	// We want the last 160 bits of the sha3-256 sum as the address
	sum := sha256.Sum256(pubkeyBytes)
	return sum[Len256AddrToLen160Addr:], nil
}
func ConvAddress(bs []byte) ([20]byte, error) {
	var addr [20]byte
	copy(addr[:], bs)
	return addr, nil
}

type marble struct {
	ObjectType string
	Name       string
	Color      string
	Size       int
	Owner      string
}

// Marshal 实现该接口后可以使用 PutKVCommon 直接传入结构体value
func (m marble) Marshal() ([]byte, error) {
	return json.Marshal(m)
}

func Unmarshal(data []byte) (interface{}, error) {
	var value marble
	err := json.Unmarshal(data, &value)
	if err != nil {
		return nil, err
	}
	return &value, nil
}

// 插入一个marble
func insertMarble(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 4 {
		return nil, errors.New("the argNum for insertMarble is not correct, expected 4")
	}
	marbleName := string(args[0])
	marbleOwner := string(args[1])
	marbleColor := string(args[2])
	marbleSizeStr := string(args[3])
	marbleSize, err := strconv.Atoi(marbleSizeStr)
	if err != nil {
		return nil, errors.New("the marble size is not int type")
	}

	marbleInfo := &marble{Name: marbleName, Color: marbleColor,
		Size: marbleSize, Owner: marbleOwner, ObjectType: "marble"}
	err = stub.PutKVCommon(marbleInfo.Name, marbleInfo)
	if err != nil {
		return nil, errors.New(err.Error())
	}

	return nil, nil
}

// 查询一个marble
func queryMarble(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for queryMarble is not correct, expected 1")
	}

	key := string(args[0])
	value, err := stub.GetKV(key)
	if err != nil {
		errInfo := fmt.Sprintf("get the key: %s failed", key)
		return nil, errors.New(errInfo)
	}

	return value, nil
}

// 删除一个marble
func deleteMarble(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for deleteMarble is not correct, expected 1")
	}

	key := string(args[0])
	value, err := stub.GetKV(key)
	if err != nil {
		errInfo := fmt.Sprintf("get the marble info err:%s", err.Error())
		return nil, errors.New(errInfo)
	}
	if value == nil {
		return nil, errors.New("the key to be delete is not exist")
	}
	err = stub.DelKV(key)
	if err != nil {
		return nil, err
	}

	return nil, nil
}

// 查询某个key的历史
func queryMarbleKeyHistory(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for queryMarbleKeyHistory is not correct, expected 1")
	}

	key := string(args[0])

	iterator, err := stub.GetKeyHistoryIterator(key)
	if err != nil {
		return nil, err
	}
	defer iterator.Close()

	var historyArray []keyHistory
	for iterator.Next() {
		var history keyHistory
		history.Value = string(iterator.Value())
		history.TxHash = iterator.TxHash()
		history.BlockNum, history.TxNum = iterator.Version()
		history.IsDeleted = iterator.IsDeleted()
		history.Timestamp = iterator.Timestamp()
		historyArray = append(historyArray, history)
	}

	historyMapBytes, err := json.Marshal(historyArray)
	if err != nil {
		return nil, err
	}

	return historyMapBytes, nil
}

// 查询指定范围的记录
func queryMarbleInRange(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 2 {
		return nil, errors.New("the argNum for queryMarbleInRange is not correct, expected 2")
	}
	begin, end := string(args[0]), string(args[1])

	iterator, err := stub.GetIterator(begin, end)
	if err != nil {
		return nil, err
	}
	defer iterator.Close()

	// 将指定范围的记录全部放入map
	rangeMap := make(map[string]string)
	var count = 0
	for iterator.Next() {
		key := iterator.Key()
		value := string(iterator.Value())
		rangeMap[key] = value
		count++
	}
	rangeMapBytes, err := json.Marshal(rangeMap)
	if err != nil {
		return nil, err
	}

	return rangeMapBytes, nil
}

// 插入marble的同时建立索引
func insertMarbleAndIndex(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 4 {
		return nil, errors.New("the args number for insertMarbleAndIndex is not correct, expected 4")
	}
	marbleName := string(args[0])
	marbleOwner := string(args[1])
	marbleColor := string(args[2])
	marbleSizeStr := string(args[3])
	marbleSize, err := strconv.Atoi(marbleSizeStr)
	if err != nil {
		return nil, errors.New("the marble size is not int type")
	}

	marbleInfo := &marble{Name: marbleName, Color: marbleColor, Size: marbleSize,
		Owner: marbleOwner, ObjectType: "marble"}
	err = stub.PutKVCommon(marbleName, marbleInfo)
	if err != nil {
		return nil, err
	}

	err = stub.SaveComIndex(indexName, []string{marbleColor, marbleSizeStr}, marbleName)
	if err != nil {
		return nil, errors.New("SaveComIndex failed")
	}

	return nil, nil
}

// 通过索引查询marbles列表
func queryMarblesByIndex(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 2 {
		return nil, errors.New("the argNum for queryMarblesByIndex is not correct, expected 2")
	}

	color := string(args[0])
	size := string(args[1])
	// 查询所有指定属性的marble
	kvIterator, err := stub.GetKVByComIndex(indexName, []string{color, size})
	if err != nil {
		return nil, err
	}
	defer kvIterator.Close()
	var array []map[string]string
	for kvIterator.Next() {
		key := kvIterator.Key()
		value := string(kvIterator.Value())
		array = append(array, map[string]string{key: value})
	}
	marshal, err := json.Marshal(array)
	if err != nil {
		return nil, err
	}
	return marshal, nil
}

// 通过索引转移marbles
func transferMarblesByIndex(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 3 {
		return nil, errors.New("the argNum for transferMarblesByIndex is not correct, expected 3")
	}

	color := string(args[0])
	size := string(args[1])
	newOwner := strings.ToLower(string(args[2]))

	// 查询所有指定属性的marble
	kvIterator, err := stub.GetKVByComIndex(indexName, []string{color, size})
	if err != nil {
		return nil, err
	}
	defer kvIterator.Close()
	for kvIterator.Next() {
		key := kvIterator.Key()
		value := kvIterator.Value()
		marbleToTransfer := marble{}
		err = json.Unmarshal(value, &marbleToTransfer)
		if err != nil {
			return nil, errors.New("Unmarshal marble failed")
		}
		marbleToTransfer.Owner = newOwner

		err = stub.PutKVCommon(key, marbleToTransfer)
		if err != nil {
			return nil, err
		}
	}

	return nil, nil
}

// 删除marble和它对应的复合索引
func deleteMarbleAndIndex(stub contractapi.ContractStub, args [][]byte) ([]byte, error) {
	if len(args) != 1 {
		return nil, errors.New("the argNum for deleteMarbleAndIndex is not correct, expected 1")
	}

	key := string(args[0])
	// 删除前检查
	value, err := stub.GetKV(key)
	if err != nil {
		errInfo := fmt.Sprintf("get the marble info err:%s", err.Error())
		return nil, errors.New(errInfo)
	}
	if value == nil {
		return nil, errors.New("the key to be delete is not exist")
	}
	// 删除key
	err = stub.DelKV(key)
	// 反序列化获取对应复合索引的值
	unmarshal, err := Unmarshal(value)
	if err != nil {
		return nil, errors.New("unmarshal value failed")
	}
	m, ok := unmarshal.(*marble)
	if !ok {
		return nil, errors.New("the value is not marble struct")
	}
	// 删除对应的复合索引值
	err = stub.DelComIndexOneRow(indexName, []string{m.Color, strconv.Itoa(m.Size)}, m.Name)
	if err != nil {
		return nil, err
	}

	return nil, nil
}
